import base64
import zlib

from .data import RESOURCES, TEMPLATES


def _unpack_payload(x: bytes) -> str:
    return zlib.decompress(base64.b64decode(x)).decode("utf-8")


def get_resource_str(template_name: str) -> str | None:
    if t := RESOURCES.get(template_name):
        return _unpack_payload(t)
    return None


def get_template_str(template_name: str) -> str | None:
    if t := TEMPLATES.get(template_name):
        return _unpack_payload(t)
    return None



# NOTE: This file is auto-generated. DO NOT EDIT!
# Update by running the __main__.py alongside this file

from typing import Final


RESOURCES: Final = {
    "_ruyi_completion": b"eNqlVntP40YQ/z+fYmosCC0h5KpKd7Tmjoc5ogYSxUlPd4FaTryJV9i7kXdNLofoZ++M7TgmmOiutVDYncdvZuexOzsTGc19NoU4WfJabQc6/J6FS/BAxzyKmA++XAh4YLHiUoCcQqD1XB03mzOug2R8iPrNe67CZXLf9OIZwYVMs+Y4lONm5HHxnOqpwM13iHfoN935UgdSNEpS6MUVC+doEgLvgYHPleZiomEesyn/CtNYRqADBjLmMy68ENQk5nMNWoL3ILkPwosYTGQYcvJa0bn6iQAlI6YDLmYHECUIOQOZ6HmiEQhi5vOYTVIqT6HIgs/GyQyUjpkXIYjP5kz4qaJI+Q9emDCKinvaP3cv7LPhx0OUa0/htP/xvHvd69gD2x06tjuwr3uX7Y7tAFegmD6ARDHQLJpPecgUTNGHdu/8sObmEXFLEXEpO/gj6vvwWAP8+BRGI2h8A8N8fNVU48mAu7vfyVORatG3Fd/lQrAYQT8YhULMdBJn+lOe/gvlBIOuM9ctw6xH93SQ/UzHLfvjDC66w4FL7tycXtsonKsZP+ZIZnIifWaZ71PKxNOwRktJcfSCkvoOJinWnmq177BZGeJ1eitjSm7C25PdFrw72X0DrZOmzx6aIglDeEPUP4p9qsJCxV7X3X1XqYSxpwO84n/hdRao9qVjmXu3R61f90rUrE0swyifT9D5vjhX7l9232l3byoPuAr/qnFVwSnRrLpZJ8OGib8G3BYy9FFFuJ32DZXA2fDy0u5XSvS67ZsBipwP+073hUi5tKzWFqbrXNmdjmV8U4GxVWzY6/Vtx3Gd3ul5BeS2cgHzMQvoccN8XMjYV6PW3dPT/v5G1IRUMtZWvYI+9yaszMCUcNXwdCNkntLw2+HbjTzQt8JryHy51s/7c51dTG0pQZRa6x+oj/6+vb3bP4bdXRKIPD0JyPWUObKax3cvCyA3nDnccGBvr9Kq6zOKyZgRcBETo1wlxMn8Hn0gVrYl4HT/skGyaG1k6qh8UEKXeP//Ilc+YgdB0T4VR9nMe9VZqEj6dq/z+bvrulhvL+1iUyk2+NwrsGi9tX5TqU/d/sVZ3z7901nprSn/u4Poxf6RFjKf7/9rQ7XKfZTVsvkefrLgqLo2E4Fv6jplBY+FufKGW5YFLUSiFsjapFBtlPrArDa2qrdGUW+b9ZNe1uUHZNsFuwpECY+WOJV5SZhaoSQU20t4LYjZEFc0zw58QhP4UIYhznHjZTqxfFFBqRtBLRU+3QfIwrFkwfHBwgEHFzrIIYxQer6XaDlNxMSABSFywTVHVBwTkUECiO+hnqGCTCz0NL7eUhyARKPxgit2kAOSE/m8tvDwNmA4QxkNBFAyHYmMfCI1aAaL2QyHP4TCkYycIu0c58UhUvrKofI9WnrtTHwQXLKIQ6jQ7KumQviZfK4YlF6LcjGUPLumVoP09uTkxYF//wJg/FlP",  # fmt: skip
    "binfmt.conf.jinja": b"eNpNUrlu20AQ7f0VAwgCbECk+wQp06Zwl0oakUNyoD2UvQRC1r/7LUkhrkjuzjuHOzqrG2w6Wo0ddd4NOlLMmvhshAYfKE0a6SPPSkVDymxIHN68s+LS4WVH6ijOMYntm5XrSQO05fSTogidLLtNqu1PC3EvidXEFhR/faY4+Wx6CjIqyAJ5J+QHyFcbxvibuhGYznDgpN69xrdD1Wa68QwOvl6DvwbltPqefYYGuIKecwXQO5yEop0QzPAIjXeS1B0I0+qKvwhoqp7YbDjhFIY5zAtVFDOgAV4GTgF9NP8wd6JbgLKE9uW+X3ThYqy+gkRvivTHehBp/wD57yfx/b7MtfB3NTwfHVuhxwMcDemw3qHmBQXcH49QaeJUrRCb6GnigiNfR4v29VaifF8NFUYZ2GIk4EJ2Sa1s+/KhR3qguZtUQNT5EKRLW/JaVhSUlLSLP+Bgi3Y5UKnwp71WsfX4+gabO2pqqAtC/MKz0CcWuiaCqb6i92s8fCFhU7+2Er79gUes69nDf9gX9HfteA==",  # fmt: skip
    "meson-cross.ini.jinja": b"eNptkMFqwzAQRO/6ioUQ0tJazaU9FPoLvfUUglDkNV4iS0IrB4eQf69k2ZRCbmI083Z3NvDDCJbO+Ck2YgMDsnfAmMYATWOiZ246sgi3G1zQXVT0PsH9/jYbq0GSI5BSzoBvxJYXzF6+f8j9ixTicCKnIyEfhYEv2GWaMRmzEyaEVZimWbltG+h8hDNeXyHo1AO5ClS6bSmRd9qqFSgp4cBPz7C9iwzJoQxZiHN4RaJrC7XYtkAd8JXnW7JwOI1kU5PH+FDwecsQsaNp4azWiiqknC+5EH3AmOpdgz6jSt5b02tyam5tuezR17rX/1Xys1b8aHLzN1oces9JDdr05PBYcrmHkrLkxqn0OqpOD2SvaxnRG2TOFdTax4d6GaBd5aRkcSd+ATx1t64=",  # fmt: skip
    "prompt.venv-created.txt.jinja": b"eNp1kTFvwkAMhff7FV7YStjbqWLqwMZSISSOO4dYJOfIvoRGiP9eXwR0KTedbOv5e8/bBmEkyYNvAZP9OHWYMpBC4gsEQZ8xVs598wCdn8CHTKPVgDIcJ1AeJFA6QTYh3/fCvVBp3+eIE2gQ6k0ylSG3Owli2h8p7VZ7iCQYMsv0Bj5FiPjUN3FKI5+L+EGGiZZ/zUPlroslUA06qTBnWNyc+4RaUJtHbdUL1vRTrPhWGYxtJDUgjHeW/5xX7mt270dPrT+25iTPszW3LV8KTe9z8+4c2LubuV6fILeb2ZrpMEUDLGDbFyHPWIE7VLhQbsDCjtOy8xFhvfFnhMzchsYbbU0tupLQBtUiDcKqc7GCdYPh/MoOzFA1iw2w4oezdaWujwiKhhpEymXNoP5kp02aZQjleFq5Xz21xdY=",  # fmt: skip
    "ruyi-activate.bash.jinja": b"eNp9lN1u2kAQhe/3KSYGqiSVRUnuqIhEGqRESgLCBClKI2uxx/JKZh2t126o6yfqI/QuT9bxH2CHhBvEzpn1d2YO7sDCFxF4IkBYx5GGFUIcoQu/hPbBiMJYOQgrIfsq3giTO1okXKMBp54K17DikX/KOrAJY3C4lKEGFUsQGlyh0NHBhjEmPHgCo5tejq1r25o+zH9MzMyAEZ19M+D5O2gfJQP6oOOHYDzSZQVL9XSdE0aOEi96CD+79XHefPHlrGx8pUeenzNPMFaAulijwvEJpIWoAwoj1BAGLqBMhArlGqWGhCvBVwFGleqIcM3f0E2X4/lXO4Nn6tOxkhFoFSOQHSoAMbnoBFzRsLgGHgRFO1WrfnJszx8eb+zp7ZU9Gy+uh3QZ+YU9w/knL42MblNrbMv4+hIqXai2Z7HMfTQ7imLuvzQhZMID4eYDoKcV7bQhx8ei7tPewFRwdtF3MenLmOg/xbcGH8Bbgya7NXiPbg0+Jq9qNXhZL8rLyf3y0Jk9m0/vZosa94mAiXVQRUqG0sWIFpWvH1vEHbAw8KAWHLWwTA9a4anZMkYhL0VCKQww4Y3gtDPXpGBsCz9K050T+AO0hixj1aB2tllztzTiMhTVj62wT//MYVXbDwr7LAAHl5/28mHuhnw/vpuA2ctYe+4NB6WstpH2TMAgwsN9RveYXhco+Rphz4JxYpSN0iWA3vth1Psu3yNlMnfFqxtrfHk7qURmO6ONpBFCSl8kYtvsvv2d0+5g/8ryquztHzTke2mmRPwHax2J0Q==",  # fmt: skip
    "ruyi-cache.toml.jinja": b"eNptUs1uhCAYvPsUX2nMtknjoccmve0e9tK9eDOGsIJK1r8CmhjjuxcUF6uexG+GmWHgFX5u4eULwpxLSHnBQH9LUpGMUbj3INqeB3C+GRpcztfwxfOihCQ5o7j7jL3BB57CLytbfOcV+ONz+Q1oGBwyjkhzWUU13R/tvkbUxhMbEqs6AzjxYINOZmkt4PEBHWjJDR5wxUr59q5VtPFDO9oMnTH3ZnezX+P/k6xFFREZU3Iltkq0gFY/9lRdF0lOeGXOSLlYHKPTFjnFUwo3TguSySP+BFj6XFMXIQfLXoq6VijW0XbTIz0LzYr7O9DiWZJgXklFigLrqJP0ZuaEN8Ch7LPobbtJSXHJFKFEEVyS5rjmHQvZwlHsUSZXpzR/S7PT3WDVNvoRuxpWU9epy/cHatoHPQ==",  # fmt: skip
    "ruyi-venv.toml.jinja": b"eNqtkOFqwyAQx7/7FBIIbWEN7AH2JEHCYUwqbTxRN1ZC3n0atQlpSxnbN/V+3v3uX3NUnewZ0QY7eRH0gxbjSPNtmgoyllR21F6tQXT0WE4knyObb5E9UqFaz3uM1INw0IIDRr6EsRKV//JOQscBuEGqz32Tmb0+zM2N0NjINovUu/SwY/MEDk70aK5LPb8kQMEglmK4pcKikGrpIZWTehQL9nHvrFcXac/GSxcsELf1Kg38DL2wVWIYCRPWyz1pc/Cjw6Als+1UMXxewKEJvI1zPdKhoQ5ML1wDhp/eqKZSPf9VSScGu48BP9DOeBWSWfUNwdzvolfWQaR8tcK3M3Dvv1VeU7PnI9EAsb8rOcQLP4FUv4t1++1Vrjf+X4L9AYyLQa0=",  # fmt: skip
    "toolchain.cmake.jinja": b"eNqFkVFrwjAUhd/7Ky6IMGF274M9dG2cZWlT0joUhJDVaIuajKQTR+l/X6zOWccwb+F+5+Scmx5MjIBNuRaPTs/pQb7lawGu68Lc3g5nGPiR94pYRgj2x14Ys1GI0VNdw07IHdNKVdA0D5VSm7zgpXSPFnOn7g+hXIL5Mi3Tb64MwzjNPIxZQtEonB4Mf9CmOcmFXFiHs9LGchwjqrujQTpLMxSx2IsQ4FJ+7gd/hwklPkpTQsH6f2iVC2OUti9csj7zSZTYVi2V59fj6bQL7PctUfcv+w1tzF/NKIwDRgnJWOJlY+iUa5UX1YzgOi9gaXMVgi+ENsDlwn7Ku+a6FAZKCVUhoOJ6JSqr3JVaya2Q1X1XbfutNN+eBYUyHfzffCwiATrs6oV6EcToDdHBDRiHz9SjMyAxnt1iw9jHkwCd2G+4PsQk",  # fmt: skip
}

TEMPLATES: Final = {
    "binfmt.conf": RESOURCES["binfmt.conf.jinja"],
    "meson-cross.ini": RESOURCES["meson-cross.ini.jinja"],
    "prompt.venv-created.txt": RESOURCES["prompt.venv-created.txt.jinja"],
    "ruyi-activate.bash": RESOURCES["ruyi-activate.bash.jinja"],
    "ruyi-cache.toml": RESOURCES["ruyi-cache.toml.jinja"],
    "ruyi-venv.toml": RESOURCES["ruyi-venv.toml.jinja"],
    "toolchain.cmake": RESOURCES["toolchain.cmake.jinja"],
}



import glob
import os
from os import PathLike
import pathlib
import re
import shutil
from typing import Any, Final, Iterator, TypedDict

from ...config import GlobalConfig
from ...log import RuyiLogger, humanize_list
from ...ruyipkg.atom import Atom
from ...ruyipkg.pkg_manifest import BoundPackageManifest, EmulatorProgDecl
from ...ruyipkg.profile import ProfileProxy
from ...utils.global_mode import ProvidesGlobalMode
from ...utils.templating import render_template_str
from .emulator_cfg import ResolvedEmulatorProg


class ConfiguredTargetTuple(TypedDict):
    target: str
    toolchain_root: PathLike[Any]
    toolchain_sysroot: PathLike[Any] | None
    toolchain_flags: str
    binutils_flavor: str
    cc_flavor: str
    gcc_install_dir: PathLike[Any] | None


class VenvPackageInfo(TypedDict):
    repo_id: str
    category: str
    name: str
    version: str


class VenvMetadata(TypedDict):
    emulator_pkgs: dict[str, VenvPackageInfo]
    extra_pkgs: list[VenvPackageInfo]
    sysroot_pkg: VenvPackageInfo | None
    toolchain_pkgs: dict[str, VenvPackageInfo]


def _venv_pkg_info_from_pkg(pkg: BoundPackageManifest) -> VenvPackageInfo:
    return VenvPackageInfo(
        repo_id=pkg.repo_id,
        category=pkg.category,
        name=pkg.name,
        version=pkg.ver,
    )


def do_make_venv(
    config: GlobalConfig,
    host: str,
    profile_name: str,
    dest: pathlib.Path,
    with_sysroot: bool,
    override_name: str | None = None,
    tc_atoms_str: list[str] | None = None,
    emu_atom_str: str | None = None,
    sysroot_atom_str: str | None = None,
    extra_cmd_atoms_str: list[str] | None = None,
) -> int:
    logger = config.logger

    # TODO: support omitting this if user only has one toolchain installed
    # this should come after implementation of local state cache
    if tc_atoms_str is None:
        logger.F(
            "You have to specify at least one toolchain atom for now, e.g. [yellow]`-t gnu-plct`[/]"
        )
        return 1

    mr = config.repo

    profile = mr.get_profile(profile_name)
    if profile is None:
        logger.F(f"profile '{profile_name}' not found")
        return 1

    target_arch = ""
    seen_target_tuples: set[str] = set()
    targets: list[ConfiguredTargetTuple] = []
    warn_differing_target_arch = False
    venv_metadata = VenvMetadata(
        emulator_pkgs={},
        extra_pkgs=[],
        sysroot_pkg=None,
        toolchain_pkgs={},
    )

    for tc_atom_str in tc_atoms_str:
        tc_atom = Atom.parse(tc_atom_str)
        tc_pm = tc_atom.match_in_repo(mr, config.include_prereleases)
        if tc_pm is None:
            logger.F(f"cannot match a toolchain package with [yellow]{tc_atom_str}[/]")
            return 1

        if tc_pm.toolchain_metadata is None:
            logger.F(f"the package [yellow]{tc_atom_str}[/] is not a toolchain")
            return 1

        if not tc_pm.toolchain_metadata.satisfies_quirk_set(profile.need_quirks):
            logger.F(
                f"the package [yellow]{tc_atom_str}[/] does not support all necessary features for the profile [yellow]{profile_name}[/]"
            )
            logger.I(
                f"quirks needed by profile:   {humanize_list(profile.need_quirks, item_color='cyan')}"
            )
            logger.I(
                f"quirks provided by package: {humanize_list(tc_pm.toolchain_metadata.quirks, item_color='yellow')}"
            )
            return 1

        target_tuple = tc_pm.toolchain_metadata.target
        if target_tuple in seen_target_tuples:
            logger.F(
                f"the target tuple [yellow]{target_tuple}[/] is already covered by one of the requested toolchains"
            )
            logger.I(
                "for now, only toolchains with differing target tuples can co-exist in one virtual environment"
            )
            return 1

        toolchain_root = config.lookup_binary_install_dir(
            host,
            tc_pm.name_for_installation,
        )
        if toolchain_root is None:
            logger.F("cannot find the installed directory for the toolchain")
            return 1

        tc_sysroot_dir: PathLike[Any] | None = None
        gcc_install_dir: PathLike[Any] | None = None
        if with_sysroot:
            if tc_sysroot_relpath := tc_pm.toolchain_metadata.included_sysroot:
                tc_sysroot_dir = pathlib.Path(toolchain_root) / tc_sysroot_relpath
                venv_metadata["sysroot_pkg"] = _venv_pkg_info_from_pkg(tc_pm)
            else:
                if sysroot_atom_str is None:
                    logger.F(
                        "sysroot is requested but the toolchain package does not include one, and [yellow]--sysroot-from[/] is not given"
                    )
                    return 1

                # try extracting from the sysroot package
                # for now only GCC toolchain packages can provide sysroots, so this is
                # okay
                gcc_pkg_atom = Atom.parse(sysroot_atom_str)
                gcc_pkg_pm = gcc_pkg_atom.match_in_repo(mr, config.include_prereleases)
                if gcc_pkg_pm is None:
                    logger.F(
                        f"cannot match a toolchain package with [yellow]{sysroot_atom_str}[/]"
                    )
                    return 1

                if gcc_pkg_pm.toolchain_metadata is None:
                    logger.F(
                        f"the package [yellow]{sysroot_atom_str}[/] is not a toolchain"
                    )
                    return 1

                gcc_pkg_root = config.lookup_binary_install_dir(
                    host,
                    gcc_pkg_pm.name_for_installation,
                )
                if gcc_pkg_root is None:
                    logger.F(
                        "cannot find the installed directory for the sysroot package"
                    )
                    return 1

                tc_sysroot_relpath = gcc_pkg_pm.toolchain_metadata.included_sysroot
                if tc_sysroot_relpath is None:
                    logger.F(
                        f"sysroot is requested but the package [yellow]{sysroot_atom_str}[/] does not contain one"
                    )
                    return 1

                tc_sysroot_dir = pathlib.Path(gcc_pkg_root) / tc_sysroot_relpath

                # also figure the GCC include/libs path out for Clang to be able to
                # locate them
                gcc_install_dir = find_gcc_install_dir(
                    gcc_pkg_root,
                    # we should use the GCC-providing package's target tuple as that's
                    # not guaranteed to be the same as llvm's
                    gcc_pkg_pm.toolchain_metadata.target,
                )

                # for now, require this directory to be present (or clang would barely work)
                if gcc_install_dir is None:
                    logger.F(
                        "cannot find a GCC include & lib directory in the sysroot package"
                    )
                    return 1

                venv_metadata["sysroot_pkg"] = _venv_pkg_info_from_pkg(gcc_pkg_pm)

        # derive flags for (the quirks of) this toolchain
        tc_flags = profile.get_common_flags(tc_pm.toolchain_metadata.quirks)

        # record the target tuple info to configure in the venv
        configured_target: ConfiguredTargetTuple = {
            "target": target_tuple,
            "toolchain_root": toolchain_root,
            "toolchain_sysroot": tc_sysroot_dir,
            "toolchain_flags": tc_flags,
            # assume clang is preferred if package contains clang
            # this is mostly true given most packages don't contain both
            "cc_flavor": "clang" if tc_pm.toolchain_metadata.has_clang else "gcc",
            # same for binutils provider flavor
            "binutils_flavor": (
                "llvm" if tc_pm.toolchain_metadata.has_llvm else "binutils"
            ),
            "gcc_install_dir": gcc_install_dir,
        }
        logger.D(f"configuration for {target_tuple}: {configured_target}")
        targets.append(configured_target)
        seen_target_tuples.add(target_tuple)
        venv_metadata["toolchain_pkgs"][target_tuple] = _venv_pkg_info_from_pkg(tc_pm)

        # record the target architecture for use in emulator package matching
        if not target_arch:
            target_arch = tc_pm.toolchain_metadata.target_arch
        elif target_arch != tc_pm.toolchain_metadata.target_arch:
            # first one wins
            warn_differing_target_arch = True

    if warn_differing_target_arch:
        logger.W("multiple toolchains specified with differing target architecture")
        logger.I(
            f"using the target architecture of the first toolchain: [yellow]{target_arch}[/]"
        )

    # Now handle the emulator.
    emu_progs = None
    emu_root: PathLike[Any] | None = None
    if emu_atom_str:
        emu_atom = Atom.parse(emu_atom_str)
        emu_pm = emu_atom.match_in_repo(mr, config.include_prereleases)
        if emu_pm is None:
            logger.F(f"cannot match an emulator package with [yellow]{emu_atom_str}[/]")
            return 1

        if emu_pm.emulator_metadata is None:
            logger.F(f"the package [yellow]{emu_atom_str}[/] is not an emulator")
            return 1

        emu_progs = list(emu_pm.emulator_metadata.list_for_arch(target_arch))
        if not emu_progs:
            logger.F(
                f"the emulator package [yellow]{emu_atom_str}[/] does not support the target architecture [yellow]{target_arch}[/]"
            )
            return 1

        for prog in emu_progs:
            if not profile.check_emulator_flavor(
                prog.flavor,
                emu_pm.emulator_metadata.quirks,
            ):
                logger.F(
                    f"the package [yellow]{emu_atom_str}[/] does not support all necessary features for the profile [yellow]{profile_name}[/]"
                )
                logger.I(
                    f"quirks needed by profile:   {humanize_list(profile.get_needed_emulator_pkg_flavors(prog.flavor), item_color='cyan')}"
                )
                logger.I(
                    f"quirks provided by package: {humanize_list(emu_pm.emulator_metadata.quirks or [], item_color='yellow')}"
                )
                return 1

        emu_root = config.lookup_binary_install_dir(
            host,
            emu_pm.name_for_installation,
        )
        if emu_root is None:
            logger.F("cannot find the installed directory for the emulator")
            return 1

        venv_metadata["emulator_pkgs"][target_arch] = _venv_pkg_info_from_pkg(emu_pm)

    # Now resolve extra commands to provide in the venv.
    extra_cmds: dict[str, str] = {}
    if extra_cmd_atoms_str:
        for extra_cmd_atom_str in extra_cmd_atoms_str:
            extra_cmd_atom = Atom.parse(extra_cmd_atom_str)
            extra_cmd_pm = extra_cmd_atom.match_in_repo(
                mr,
                config.include_prereleases,
            )
            if extra_cmd_pm is None:
                logger.F(
                    f"cannot match an extra command package with [yellow]{extra_cmd_atom_str}[/]"
                )
                return 1

            extra_cmd_bm = extra_cmd_pm.binary_metadata
            if not extra_cmd_bm:
                logger.F(
                    f"the package [yellow]{extra_cmd_atom_str}[/] is not a binary-providing package"
                )
                return 1

            extra_cmds_decl = extra_cmd_bm.get_commands_for_host(host)
            if not extra_cmds_decl:
                logger.W(
                    f"the package [yellow]{extra_cmd_atom_str}[/] does not provide any command for host [yellow]{host}[/], ignoring"
                )
                continue

            cmd_root = config.lookup_binary_install_dir(
                host,
                extra_cmd_pm.name_for_installation,
            )
            if cmd_root is None:
                logger.F(
                    f"cannot find the installed directory for the package [yellow]{extra_cmd_pm.name_for_installation}[/]"
                )
                return 1
            cmd_root = pathlib.Path(cmd_root)

            venv_metadata["extra_pkgs"].append(_venv_pkg_info_from_pkg(extra_cmd_pm))

            for cmd, cmd_rel_path in extra_cmds_decl.items():
                # resolve the command path
                cmd_path = (cmd_root / cmd_rel_path).resolve()
                if not cmd_path.is_relative_to(cmd_root):
                    # we don't allow commands to resolve outside of the
                    # providing package's install root
                    logger.F(
                        "internal error: resolved command path is outside of the providing package"
                    )
                    return 1

                # add the command to the list
                extra_cmds[cmd] = str(cmd_path)

    if override_name is not None:
        logger.I(
            f"Creating a Ruyi virtual environment [cyan]'{override_name}'[/] at [green]{dest}[/]..."
        )
    else:
        logger.I(f"Creating a Ruyi virtual environment at [green]{dest}[/]...")

    maker = VenvMaker(
        config,
        profile,
        targets,
        dest.resolve(),
        emu_progs,
        emu_root,
        extra_cmds,
        venv_metadata,
        override_name,
    )
    maker.provision()

    logger.I(
        render_template_str(
            "prompt.venv-created.txt",
            {
                "sysroot": maker.sysroot_destdir(None),
            },
        )
    )

    return 0


def find_gcc_install_dir(
    install_root: PathLike[Any],
    target_tuple: str,
) -> PathLike[Any] | None:
    # check $PREFIX/lib/gcc/$TARGET/*
    search_root = pathlib.Path(install_root) / "lib" / "gcc" / target_tuple
    try:
        for p in search_root.iterdir():
            # only want the first one (should be the only one)
            return p
    except FileNotFoundError:
        pass

    # nothing?
    return None


class VenvMaker:
    """Performs the actual creation of a Ruyi virtual environment."""

    def __init__(
        self,
        gc: GlobalConfig,
        profile: ProfileProxy,
        targets: list[ConfiguredTargetTuple],
        dest: PathLike[Any],
        emulator_progs: list[EmulatorProgDecl] | None,
        emulator_root: PathLike[Any] | None,
        extra_cmds: dict[str, str] | None,
        metadata: VenvMetadata,
        override_name: str | None = None,
    ) -> None:
        self.gc = gc
        self.profile = profile
        self.targets = targets
        self.venv_root = pathlib.Path(dest)
        self.emulator_progs = emulator_progs
        self.emulator_root = emulator_root
        self.extra_cmds = extra_cmds or {}
        self.metadata = metadata
        self.override_name = override_name

        self.bindir = self.venv_root / "bin"

    @property
    def logger(self) -> RuyiLogger:
        return self.gc.logger

    def render_and_write(
        self,
        dest: PathLike[Any],
        template_name: str,
        data: dict[str, Any],
    ) -> None:
        self.logger.D(f"rendering template '{template_name}' with data {data}")
        content = render_template_str(template_name, data).encode("utf-8")
        self.logger.D(f"writing {dest}")
        with open(dest, "wb") as fp:
            fp.write(content)

    def sysroot_srcdir(self, target_tuple: str | None) -> pathlib.Path | None:
        if target_tuple is None:
            # check the primary target
            if s := self.targets[0]["toolchain_sysroot"]:
                return pathlib.Path(s)
            return None

        # check if we have this target
        for t in self.targets:
            if t["target"] != target_tuple:
                continue
            if s := t["toolchain_sysroot"]:
                return pathlib.Path(s)

        return None

    def has_sysroot_for(self, target_tuple: str | None) -> bool:
        return self.sysroot_srcdir(target_tuple) is not None

    def sysroot_destdir(self, target_tuple: str | None) -> pathlib.Path | None:
        if not self.has_sysroot_for(target_tuple):
            return None

        dirname = f"sysroot.{target_tuple}" if target_tuple is not None else "sysroot"
        return self.venv_root / dirname

    def provision(self) -> None:
        venv_root = self.venv_root
        bindir = self.bindir

        venv_root.mkdir()
        bindir.mkdir()

        env_data = {
            "profile": self.profile.id,
            "sysroot": self.sysroot_destdir(None),
            "metadata": self.metadata,
        }
        self.render_and_write(
            venv_root / "ruyi-venv.toml",
            "ruyi-venv.toml",
            env_data,
        )

        for i, tgt in enumerate(self.targets):
            is_primary = i == 0
            self.provision_target(tgt, is_primary)

        if self.extra_cmds:
            symlink_binaries(
                self.gc,
                self.logger,
                bindir,
                src_cmds_names=list(self.extra_cmds.keys()),
            )

        template_data = {
            "RUYI_VENV": str(venv_root),
            "RUYI_VENV_NAME": self.override_name,
        }

        self.render_and_write(
            bindir / "ruyi-activate",
            "ruyi-activate.bash",
            template_data,
        )

        qemu_bin: PathLike[Any] | None = None
        profile_emu_env: dict[str, str] | None = None
        if self.emulator_root is not None and self.emulator_progs:
            resolved_emu_progs = [
                ResolvedEmulatorProg.new(
                    p,
                    self.emulator_root,
                    self.profile,
                    self.sysroot_destdir(None),
                )
                for p in self.emulator_progs
            ]
            binfmt_data = {
                "resolved_progs": resolved_emu_progs,
            }
            self.render_and_write(
                venv_root / "binfmt.conf",
                "binfmt.conf",
                binfmt_data,
            )

            for i, p in enumerate(self.emulator_progs):
                if not p.is_qemu:
                    continue

                qemu_bin = pathlib.Path(self.emulator_root) / p.relative_path
                profile_emu_env = resolved_emu_progs[i].env

                self.logger.D("symlinking the ruyi-qemu wrapper")
                os.symlink(self.gc.self_exe, bindir / "ruyi-qemu")

        # provide initial cached configuration to venv
        self.render_and_write(
            venv_root / "ruyi-cache.v2.toml",
            "ruyi-cache.toml",
            self.make_venv_cache_data(
                qemu_bin,
                self.extra_cmds,
                profile_emu_env,
            ),
        )

    def make_venv_cache_data(
        self,
        qemu_bin: PathLike[Any] | None,
        extra_cmds: dict[str, str],
        profile_emu_env: dict[str, str] | None,
    ) -> dict[str, object]:
        targets_cache_data: dict[str, object] = {
            tgt["target"]: {
                "toolchain_bindir": str(pathlib.Path(tgt["toolchain_root"]) / "bin"),
                "toolchain_sysroot": self.sysroot_destdir(tgt["target"]),
                "toolchain_flags": tgt["toolchain_flags"],
                "gcc_install_dir": tgt["gcc_install_dir"],
            }
            for tgt in self.targets
        }

        cmd_metadata_map = make_cmd_metadata_map(self.logger, self.targets)

        # add extra cmds that are not associated with any target
        for cmd, dest in extra_cmds.items():
            if cmd in cmd_metadata_map:
                self.logger.W(
                    f"extra command {cmd} is already provided by another package, overriding it"
                )
            cmd_metadata_map[cmd] = {
                "dest": dest,
                "target_tuple": "",
            }

        return {
            "profile_emu_env": profile_emu_env,
            "qemu_bin": qemu_bin,
            "targets": targets_cache_data,
            "cmd_metadata_map": cmd_metadata_map,
        }

    def provision_target(
        self,
        tgt: ConfiguredTargetTuple,
        is_primary: bool,
    ) -> None:
        venv_root = self.venv_root
        bindir = self.bindir
        target_tuple = tgt["target"]

        # getting the destdir this way ensures it's suffixed with the target
        # tuple
        if sysroot_destdir := self.sysroot_destdir(target_tuple):
            sysroot_srcdir = tgt["toolchain_sysroot"]
            assert sysroot_srcdir is not None

            self.logger.D(f"copying sysroot for {target_tuple}")
            shutil.copytree(
                sysroot_srcdir,
                sysroot_destdir,
                symlinks=True,
                ignore_dangling_symlinks=True,
            )

            if is_primary:
                self.logger.D("symlinking primary sysroot into place")
                primary_sysroot_destdir = self.sysroot_destdir(None)
                assert primary_sysroot_destdir is not None
                os.symlink(sysroot_destdir.name, primary_sysroot_destdir)

        self.logger.D(f"symlinking {target_tuple} binaries into venv")
        toolchain_bindir = pathlib.Path(tgt["toolchain_root"]) / "bin"
        symlink_binaries(self.gc, self.logger, bindir, src_bindir=toolchain_bindir)

        make_llvm_tool_aliases(
            self.logger,
            bindir,
            target_tuple,
            tgt["binutils_flavor"] == "llvm",
            tgt["cc_flavor"] == "clang",
        )

        # CMake toolchain file & Meson cross file
        if tgt["cc_flavor"] == "clang":
            cc_path = bindir / "clang"
            cxx_path = bindir / "clang++"
        elif tgt["cc_flavor"] == "gcc":
            cc_path = bindir / f"{target_tuple}-gcc"
            cxx_path = bindir / f"{target_tuple}-g++"
        else:
            raise NotImplementedError

        if tgt["binutils_flavor"] == "binutils":
            meson_additional_binaries = {
                "ar": bindir / f"{target_tuple}-ar",
                "nm": bindir / f"{target_tuple}-nm",
                "objcopy": bindir / f"{target_tuple}-objcopy",
                "objdump": bindir / f"{target_tuple}-objdump",
                "ranlib": bindir / f"{target_tuple}-ranlib",
                "readelf": bindir / f"{target_tuple}-readelf",
                "strip": bindir / f"{target_tuple}-strip",
            }
        elif tgt["binutils_flavor"] == "llvm":
            meson_additional_binaries = {
                "ar": bindir / "llvm-ar",
                "nm": bindir / "llvm-nm",
                "objcopy": bindir / "llvm-objcopy",
                "objdump": bindir / "llvm-objdump",
                "ranlib": bindir / "llvm-ranlib",
                "readelf": bindir / "llvm-readelf",
                "strip": bindir / "llvm-strip",
            }
        else:
            raise NotImplementedError

        cmake_toolchain_file_path = venv_root / f"toolchain.{target_tuple}.cmake"
        toolchain_file_data = {
            "cc": cc_path,
            "cxx": cxx_path,
            "processor": self.profile.arch,
            "sysroot": self.sysroot_destdir(target_tuple),
            "venv_root": venv_root,
            "cmake_toolchain_file": str(cmake_toolchain_file_path),
            "meson_additional_binaries": meson_additional_binaries,
        }
        self.render_and_write(
            cmake_toolchain_file_path,
            "toolchain.cmake",
            toolchain_file_data,
        )

        meson_cross_file_path = venv_root / f"meson-cross.{target_tuple}.ini"
        self.render_and_write(
            meson_cross_file_path,
            "meson-cross.ini",
            toolchain_file_data,
        )

        if is_primary:
            self.logger.D(
                f"making cmake & meson file symlinks to primary target {target_tuple}"
            )
            primary_cmake_toolchain_file_path = venv_root / "toolchain.cmake"
            primary_meson_cross_file_path = venv_root / "meson-cross.ini"
            os.symlink(
                cmake_toolchain_file_path.name,
                primary_cmake_toolchain_file_path,
            )
            os.symlink(meson_cross_file_path.name, primary_meson_cross_file_path)


def iter_binaries_to_symlink(
    logger: RuyiLogger,
    bindir: pathlib.Path,
) -> Iterator[pathlib.Path]:
    for filename in glob.iglob("*", root_dir=bindir):
        src_cmd_path = bindir / filename
        if not is_executable(src_cmd_path):
            logger.D(f"skipping non-executable {filename} in src bindir")
            continue

        if should_ignore_symlinking(filename):
            logger.D(f"skipping command {filename} explicitly")
            continue

        yield bindir / filename


class CmdMetadataEntry(TypedDict):
    dest: str
    target_tuple: str


def make_cmd_metadata_map(
    logger: RuyiLogger,
    targets: list[ConfiguredTargetTuple],
) -> dict[str, CmdMetadataEntry]:
    result: dict[str, CmdMetadataEntry] = {}
    for tgt in targets:
        # TODO: dedup this and provision_target
        toolchain_bindir = pathlib.Path(tgt["toolchain_root"]) / "bin"
        for cmd in iter_binaries_to_symlink(logger, toolchain_bindir):
            result[cmd.name] = {
                "dest": str(cmd),
                "target_tuple": tgt["target"],
            }
    return result


def symlink_binaries(
    gm: ProvidesGlobalMode,
    logger: RuyiLogger,
    dest_bindir: PathLike[Any],
    *,
    src_bindir: PathLike[Any] | None = None,
    src_cmds_names: list[str] | None = None,
) -> None:
    dest_binpath = pathlib.Path(dest_bindir)
    self_exe_path = gm.self_exe

    if src_bindir is not None:
        src_binpath = pathlib.Path(src_bindir)
        for src_cmd_path in iter_binaries_to_symlink(logger, src_binpath):
            filename = src_cmd_path.name

            # symlink self to dest with the name of this command
            dest_path = dest_binpath / filename
            logger.D(f"making ruyi symlink to {self_exe_path} at {dest_path}")
            os.symlink(self_exe_path, dest_path)
        return

    if src_cmds_names is not None:
        for cmd in src_cmds_names:
            # symlink self to dest with the name of this command
            dest_path = dest_binpath / cmd
            logger.D(f"making ruyi symlink to {self_exe_path} at {dest_path}")
            os.symlink(self_exe_path, dest_path)
        return

    raise ValueError(
        "internal error: either src_bindir or src_cmds_names must be provided"
    )


LLVM_BINUTILS_ALIASES: Final = {
    "addr2line": "llvm-addr2line",
    "ar": "llvm-ar",
    "as": "llvm-as",
    "c++filt": "llvm-cxxfilt",
    "gcc-ar": "llvm-ar",
    "gcc-nm": "llvm-nm",
    "gcc-ranlib": "llvm-ranlib",
    # 'gcov': 'llvm-cov',  # I'm not sure if this is correct
    "ld": "ld.lld",
    "nm": "llvm-nm",
    "objcopy": "llvm-objcopy",
    "objdump": "llvm-objdump",
    "ranlib": "llvm-ranlib",
    "readelf": "llvm-readelf",
    "size": "llvm-size",
    "strings": "llvm-strings",
    "strip": "llvm-strip",
}

CLANG_GCC_ALIASES: Final = {
    "c++": "clang++",
    "cc": "clang",
    "cpp": "clang-cpp",
    "g++": "clang++",
    "gcc": "clang",
}


def make_llvm_tool_aliases(
    logger: RuyiLogger,
    dest_bindir: PathLike[Any],
    target_tuple: str,
    do_binutils: bool,
    do_clang: bool,
) -> None:
    if do_binutils:
        make_compat_symlinks(logger, dest_bindir, target_tuple, LLVM_BINUTILS_ALIASES)
    if do_clang:
        make_compat_symlinks(logger, dest_bindir, target_tuple, CLANG_GCC_ALIASES)


def make_compat_symlinks(
    logger: RuyiLogger,
    dest_bindir: PathLike[Any],
    target_tuple: str,
    aliases: dict[str, str],
) -> None:
    destdir = pathlib.Path(dest_bindir)
    for compat_basename, symlink_target in aliases.items():
        compat_name = f"{target_tuple}-{compat_basename}"
        logger.D(f"making compat symlink: {compat_name} -> {symlink_target}")
        os.symlink(symlink_target, destdir / compat_name)


def is_executable(p: PathLike[Any]) -> bool:
    return os.access(p, os.F_OK | os.X_OK)


def should_ignore_symlinking(c: str) -> bool:
    return is_command_specific_to_ct_ng(c) or is_command_versioned_cc(c)


def is_command_specific_to_ct_ng(c: str) -> bool:
    return c.endswith("populate") or c.endswith("ct-ng.config")


VERSIONED_CC_RE: Final = re.compile(
    r"(?:^|-)(?:g?cc|c\+\+|g\+\+|cpp|clang|clang\+\+)-[0-9.]+$"
)


def is_command_versioned_cc(c: str) -> bool:
    return VERSIONED_CC_RE.search(c) is not None



from contextlib import AbstractContextManager
import enum
import json
import sys
from types import TracebackType
from typing import BinaryIO, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

if sys.version_info >= (3, 11):

    class PorcelainEntityType(enum.StrEnum):
        LogV1 = "log-v1"
        NewsItemV1 = "newsitem-v1"
        PkgListOutputV1 = "pkglistoutput-v1"
        EntityListOutputV1 = "entitylistoutput-v1"

else:

    class PorcelainEntityType(str, enum.Enum):
        LogV1 = "log-v1"
        NewsItemV1 = "newsitem-v1"
        PkgListOutputV1 = "pkglistoutput-v1"
        EntityListOutputV1 = "entitylistoutput-v1"


class PorcelainEntity(TypedDict):
    ty: PorcelainEntityType


class PorcelainOutput(AbstractContextManager["PorcelainOutput"]):
    def __init__(self, out: BinaryIO | None = None) -> None:
        self.out = sys.stdout.buffer if out is None else out

    def __enter__(self) -> "Self":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        self.out.flush()
        return None

    def emit(self, obj: PorcelainEntity) -> None:
        s = json.dumps(obj, ensure_ascii=False, separators=(",", ":"))
        self.out.write(s.encode("utf-8"))
        self.out.write(b"\n")



import ast
import builtins
import inspect
import os
import pathlib
import sys
from types import CodeType
from typing import Callable, Final, MutableMapping, NoReturn, TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing_extensions import Buffer

from .api import RuyiHostAPI
from .ctx import PluginHostContext, BasePluginLoader


class UnsandboxedModuleDict(dict[str, object]):
    def get_option(self, key: str) -> object:
        return self.get(key, None)


class UnsandboxedTrivialEvaluator:
    def eval_function(
        self,
        function: object,
        *args: object,
        **kwargs: object,
    ) -> object:
        if callable(function):
            return function(*args, **kwargs)
        raise RuntimeError(f"the Python value {function!r} is not callable")


BUILTINS_TO_EXPOSE: Final = {
    k: getattr(builtins, k)
    for k in [
        "abs",
        "any",
        "all",
        "bool",
        "bytes",
        "dict",
        "dir",
        "enumerate",
        "float",
        "getattr",
        "hasattr",
        "hash",
        "int",
        "len",
        "list",
        "max",
        "min",
        "print",
        "range",
        "repr",
        "reversed",
        "sorted",
        "str",
        "tuple",
        "type",
        "zip",
    ]
}


def _fail_helper(*args: object) -> NoReturn:
    raise RuntimeError(f"fail: {''.join(str(x) for x in args)}")


class UnsandboxedPluginHostContext(
    PluginHostContext[UnsandboxedModuleDict, UnsandboxedTrivialEvaluator]
):
    def make_loader(
        self,
        originating_file: pathlib.Path,
        module_cache: MutableMapping[str, UnsandboxedModuleDict],
        is_cmd: bool,
    ) -> BasePluginLoader[UnsandboxedModuleDict]:
        return UnsandboxedRuyiPluginLoader(self, originating_file, module_cache, is_cmd)

    def make_evaluator(self) -> UnsandboxedTrivialEvaluator:
        return UnsandboxedTrivialEvaluator()


def _is_name_private(n: str) -> bool:
    return n.startswith("_")


def _assert_name_is_public(n: str) -> None | NoReturn:
    if _is_name_private(n):
        raise RuntimeError(f"error: trying to load private name {n}")
    return None


class UnsandboxedRuyiPluginLoader(BasePluginLoader[UnsandboxedModuleDict]):
    def do_load_module(
        self,
        resolved_path: pathlib.Path,
        program: str,
        ruyi_host_bridge: Callable[[object], RuyiHostAPI],
    ) -> UnsandboxedModuleDict:
        self.host_logger.D(f"unsandboxed module load: path {resolved_path}")

        sub_loader = self.make_sub_loader(resolved_path)

        def _load_stmt_helper(
            spec: str,
            *values_to_bind: str,
            **renamed_values_to_bind: str,
        ) -> None:
            mod = sub_loader.load(spec)

            curr_frame = inspect.currentframe()
            if curr_frame is None:
                raise RuntimeError(
                    "cannot inspect the Python runtime for the current frame"
                )

            parent_frame = curr_frame.f_back
            if parent_frame is None:
                raise RuntimeError(
                    "internal error: no parent frame for load() statement"
                )

            g = parent_frame.f_locals
            for name in values_to_bind:
                _assert_name_is_public(name)
                g[name] = mod[name]
            for dst_name, src_name in renamed_values_to_bind.items():
                _assert_name_is_public(src_name)
                g[dst_name] = mod[src_name]
            return None

        code = self.source_to_code(program, resolved_path)
        mod_globals: dict[str, object] = {
            "__builtins__": BUILTINS_TO_EXPOSE,
            "fail": _fail_helper,
            "load": _load_stmt_helper,
            "ruyi_plugin_rev": ruyi_host_bridge,
        }
        # pylint: disable-next=exec-used
        exec(code, mod_globals)
        return UnsandboxedModuleDict(mod_globals)

    # intentionally follows the importlib.abc.InspectLoader protocol, for
    # easier refactoring whenever necessary.
    @staticmethod
    def source_to_code(
        data: "Buffer | str | ast.Module",
        path: "str | os.PathLike[str]" = "<string>",
    ) -> CodeType:
        mod_ast: ast.Module
        if isinstance(data, ast.Module):
            mod_ast = data
        else:  # isinstance(data, str) or isinstance(data, Buffer)
            mod_ast = ast.parse(data, path, "exec")

        # lint the module on a best-effort basis to help fight syntax feature
        # creep
        lint_module(mod_ast)

        return compile(mod_ast, path, "exec")


def lint_module(mod: ast.Module) -> None:
    if node := GatedLanguageFeaturesPass().visit(mod):
        raise RuntimeError(f"line {node.lineno}: language feature is gated")


class GatedLanguageFeaturesPass(ast.NodeVisitor):
    def visit(self, node: ast.AST) -> ast.expr | ast.stmt | None:
        return cast(ast.expr | ast.stmt | None, super().visit(node))

    def generic_visit(self, node: ast.AST) -> ast.expr | ast.stmt | None:
        """Traverses all types of nodes, bailing if non-minimal language
        features are found."""

        for _, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        if x := self.visit(item):
                            return x
            elif isinstance(value, ast.AST):
                if x := self.visit(value):
                    return x
        return None

    def visit_NamedExpr(self, node: ast.NamedExpr) -> ast.NamedExpr:
        return node

    def visit_Raise(self, node: ast.Raise) -> ast.Raise:
        return node

    def visit_Assert(self, node: ast.Assert) -> ast.Assert:
        return node

    def visit_Import(self, node: ast.Import) -> ast.Import:
        return node

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
        return node

    def visit_Try(self, node: ast.Try) -> ast.Try:
        return node

    if sys.version_info >= (3, 11):

        def visit_TryStar(self, node: ast.TryStar) -> ast.TryStar:
            return node

    def visit_With(self, node: ast.With) -> ast.With:
        return node

    def visit_Match(self, node: ast.Match) -> ast.Match:
        return node

    def visit_Yield(self, node: ast.Yield) -> ast.Yield:
        return node

    def visit_YieldFrom(self, node: ast.YieldFrom) -> ast.YieldFrom:
        return node

    def visit_Global(self, node: ast.Global) -> ast.Global:
        return node

    def visit_Nonlocal(self, node: ast.Nonlocal) -> ast.Nonlocal:
        return node

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        return node

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef
    ) -> ast.AsyncFunctionDef:
        return node

    def visit_Await(self, node: ast.Await) -> ast.Await:
        return node

    def visit_AsyncFor(self, node: ast.AsyncFor) -> ast.AsyncFor:
        return node

    def visit_AsyncWith(self, node: ast.AsyncWith) -> ast.AsyncWith:
        return node



from itertools import chain

from ..config import GlobalConfig
from ..log import RuyiLogger
from ..utils.porcelain import PorcelainOutput
from .augmented_pkg import AugmentedPkg
from .list_filter import ListFilter
from .pkg_manifest import BoundPackageManifest


def do_list(
    cfg: GlobalConfig,
    filters: ListFilter,
    verbose: bool,
) -> int:
    logger = cfg.logger

    if not filters:
        if cfg.is_porcelain:
            # we don't want to print message for humans in case of porcelain
            # mode, but we don't want to retain the old behavior of listing
            # all packages either
            return 1

        logger.F("no filter specified for list operation")
        logger.I(
            "for the old behavior of listing all packages, try [yellow]ruyi list --name-contains ''[/]"
        )
        return 1

    augmented_pkgs = list(AugmentedPkg.yield_from_repo(cfg, cfg.repo, filters))

    if cfg.is_porcelain:
        return _do_list_porcelain(augmented_pkgs)

    if not verbose:
        return _do_list_non_verbose(logger, augmented_pkgs)

    for i, ver in enumerate(chain(*(ap.versions for ap in augmented_pkgs))):
        if i > 0:
            logger.stdout("\n")

        _print_pkg_detail(logger, ver.pm, cfg.lang_code)

    return 0


def _do_list_non_verbose(
    logger: RuyiLogger,
    augmented_pkgs: list[AugmentedPkg],
) -> int:
    logger.stdout("List of available packages:\n")

    for ap in augmented_pkgs:
        logger.stdout(f"* [bold green]{ap.category}/{ap.name}[/]")
        for ver in ap.versions:
            if ver.remarks:
                comments_str = (
                    f" ({', '.join(r.as_rich_markup() for r in ver.remarks)})"
                )
            else:
                comments_str = ""
            slug_str = f" slug: [yellow]{ver.pm.slug}[/]" if ver.pm.slug else ""
            logger.stdout(f"  - [blue]{ver.pm.semver}[/]{comments_str}{slug_str}")

    return 0


def _do_list_porcelain(augmented_pkgs: list[AugmentedPkg]) -> int:
    with PorcelainOutput() as po:
        for ap in augmented_pkgs:
            po.emit(ap.to_porcelain())

    return 0


def _print_pkg_detail(
    logger: RuyiLogger,
    pm: BoundPackageManifest,
    lang_code: str,
) -> None:
    logger.stdout(f"[bold]## [green]{pm.category}/{pm.name}[/] [blue]{pm.ver}[/][/]\n")

    if pm.slug is not None:
        logger.stdout(f"* Slug: [yellow]{pm.slug}[/]")
    else:
        logger.stdout("* Slug: (none)")
    logger.stdout(f"* Package kind: {sorted(pm.kind)}")
    logger.stdout(f"* Vendor: {pm.vendor_name}")
    if upstream_ver := pm.upstream_version:
        logger.stdout(f"* Upstream version number: {upstream_ver}")
    else:
        logger.stdout("* Upstream version number: (undeclared)")
    logger.stdout("")

    sv = pm.service_level
    if sv.has_known_issues:
        logger.stdout("\nPackage has known issue(s):\n")
        for x in sv.render_known_issues(pm.repo.messages, lang_code):
            logger.stdout(x, end="\n\n")

    df = pm.distfiles
    logger.stdout(f"Package declares {len(df)} distfile(s):\n")
    for dd in df.values():
        logger.stdout(f"* [green]{dd.name}[/]")
        logger.stdout(f"    - Size: [yellow]{dd.size}[/] bytes")
        for kind, csum in dd.checksums.items():
            logger.stdout(f"    - {kind.upper()}: [yellow]{csum}[/]")

    if bm := pm.binary_metadata:
        logger.stdout("\n### Binary artifacts\n")
        for host, data in bm.data.items():
            logger.stdout(f"* Host [green]{host}[/]:")
            logger.stdout(f"    - Distfiles: {data['distfiles']}")
            if cmds := data.get("commands"):
                logger.stdout("    - Available command(s):")
                for k in sorted(cmds.keys()):
                    logger.stdout(f"        - [green]{k}[/]")

    if tm := pm.toolchain_metadata:
        logger.stdout("\n### Toolchain metadata\n")
        logger.stdout(f"* Target: [bold green]{tm.target}[/]")
        logger.stdout(f"* Quirks: {tm.quirks}")
        logger.stdout("* Components:")
        for tc in tm.components:
            logger.stdout(f'    - {tc["name"]} [bold green]{tc["version"]}[/]')



from typing import Any, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from ..cli.completer import DynamicCompleter
    from ..config import GlobalConfig


def package_completer_builder(
    cfg: "GlobalConfig",
    filters: list[Callable[[str], bool]] | None = None,
) -> "DynamicCompleter":
    # Lazy import to avoid circular dependency
    from ..ruyipkg.augmented_pkg import (
        AugmentedPkg,
    )  # pylint: disable=import-outside-toplevel
    from ..ruyipkg.list_filter import (
        ListFilter,
    )  # pylint: disable=import-outside-toplevel

    all_pkgs = list(
        AugmentedPkg.yield_from_repo(
            cfg,
            cfg.repo,
            ListFilter(),
            ensure_repo=False,
        )
    )
    if filters is not None:
        all_pkgs = [
            pkg
            for pkg in all_pkgs
            if pkg.name is not None and all(f(pkg.name) for f in filters)
        ]

    def f(prefix: str, parsed_args: object, **kwargs: Any) -> list[str]:
        return [
            pkg.name
            for pkg in all_pkgs
            if pkg.name is not None and pkg.name.startswith(prefix)
        ]

    return f



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class UpdateCommand(
    RootCommand,
    cmd="update",
    help="Update RuyiSDK repo and packages",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from . import news
        from .state import BoundInstallationStateStore

        logger = cfg.logger
        mr = cfg.repo
        mr.sync()

        # check for upgradable packages
        bis = BoundInstallationStateStore(cfg.ruyipkg_global_state, mr)
        upgradable = list(bis.iter_upgradable_pkgs(cfg.include_prereleases))

        if upgradable:
            logger.stdout(
                "\nNewer versions are available for some of your installed packages:\n"
            )
            for pm, new_ver in upgradable:
                logger.stdout(
                    f"  - [bold]{pm.category}/{pm.name}[/]: [yellow]{pm.ver}[/] -> [green]{new_ver}[/]"
                )
            logger.stdout(
                """
Re-run [yellow]ruyi install[/] to upgrade, and don't forget to re-create any affected
virtual environments."""
            )

        news.maybe_notify_unread_news(cfg, False)

        return 0



import abc
import os
import pathlib
from typing import (
    Callable,
    Final,
    Generic,
    MutableMapping,
    Protocol,
    TypeVar,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from typing_extensions import Self

from ..log import RuyiLogger
from . import api
from . import paths


ENV_PLUGIN_BACKEND_KEY: Final = "RUYI_PLUGIN_BACKEND"


class SupportsGetOption(Protocol):
    def get_option(self, key: str) -> object: ...


class SupportsEvalFunction(Protocol):
    def eval_function(
        self,
        function: object,
        *args: object,
        **kwargs: object,
    ) -> object: ...


ModuleTy = TypeVar("ModuleTy", bound=SupportsGetOption, covariant=True)
EvalTy = TypeVar("EvalTy", bound=SupportsEvalFunction, covariant=True)


class PluginHostContext(Generic[ModuleTy, EvalTy], metaclass=abc.ABCMeta):
    @staticmethod
    def new(
        host_logger: RuyiLogger,
        plugin_root: pathlib.Path,
    ) -> "PluginHostContext[SupportsGetOption, SupportsEvalFunction]":
        plugin_backend = os.environ.get("RUYI_PLUGIN_BACKEND", "")
        if not plugin_backend:
            plugin_backend = "unsandboxed"

        match plugin_backend:
            case "unsandboxed":
                return UnsandboxedPluginHostContext(host_logger, plugin_root)
            case _:
                raise RuntimeError(f"unsupported plugin backend: {plugin_backend}")

    def __init__(
        self,
        host_logger: RuyiLogger,
        plugin_root: pathlib.Path,
    ) -> None:
        self._host_logger = host_logger
        self._plugin_root = plugin_root
        # resolved path: finalized module
        self._module_cache: MutableMapping[str, ModuleTy] = {}
        # plugin id: finalized plugin module
        self._loaded_plugins: dict[str, SupportsGetOption] = {}
        # plugin id: {key: value}
        self._value_cache: dict[str, dict[str, object]] = {}

    @abc.abstractmethod
    def make_loader(
        self,
        originating_file: pathlib.Path,
        module_cache: MutableMapping[str, ModuleTy],
        is_cmd: bool,
    ) -> "BasePluginLoader[ModuleTy]":
        raise NotImplementedError

    @abc.abstractmethod
    def make_evaluator(self) -> EvalTy:
        raise NotImplementedError

    @property
    def host_logger(self) -> RuyiLogger:
        return self._host_logger

    @property
    def plugin_root(self) -> pathlib.Path:
        return self._plugin_root

    def load_plugin(self, plugin_id: str, is_cmd: bool) -> None:
        plugin_dir = paths.get_plugin_dir(plugin_id, self._plugin_root)

        loader = self.make_loader(
            plugin_dir / paths.PLUGIN_ENTRYPOINT_FILENAME,
            self._module_cache,
            is_cmd,
        )
        loaded_plugin = loader.load_this_plugin()
        self._loaded_plugins[plugin_id] = loaded_plugin

    def is_plugin_loaded(self, plugin_id: str) -> bool:
        return plugin_id in self._loaded_plugins

    def get_from_plugin(
        self,
        plugin_id: str,
        key: str,
        is_cmd_plugin: bool = False,
    ) -> object | None:
        if not self.is_plugin_loaded(plugin_id):
            self.load_plugin(plugin_id, is_cmd_plugin)

        if plugin_id not in self._value_cache:
            self._value_cache[plugin_id] = {}

        try:
            return self._value_cache[plugin_id][key]
        except KeyError:
            v = self._loaded_plugins[plugin_id].get_option(key)
            self._value_cache[plugin_id][key] = v
            return v


class BasePluginLoader(Generic[ModuleTy], metaclass=abc.ABCMeta):
    """Base class for plugin loaders loading from Ruyi repo.

    Load paths take one of the following shapes:

    * relative path: loads the path relative from the originating file's location,
      but crossing plugin boundary is not allowed
    * absolute path: similar to above, but relative to the plugin's FS root
    * `ruyi-plugin://${plugin-id}`: loads from the plugin `plugin-id` residing
      in the same repo as the originating plugin, the "entrypoint" being hard-coded
      to whatever the concrete implementation dictates
    """

    def __init__(
        self,
        phctx: PluginHostContext[ModuleTy, SupportsEvalFunction],
        originating_file: pathlib.Path,
        module_cache: MutableMapping[str, ModuleTy],
        is_cmd: bool,
    ) -> None:
        self._phctx = phctx
        self.originating_file = originating_file
        self.module_cache = module_cache
        self.is_cmd = is_cmd

    @property
    def host_logger(self) -> RuyiLogger:
        return self._phctx.host_logger

    @property
    def root(self) -> pathlib.Path:
        return self._phctx.plugin_root

    def make_sub_loader(self, originating_file: pathlib.Path) -> "Self":
        return self.__class__(
            self._phctx,
            originating_file,
            self.module_cache,
            self.is_cmd,
        )

    def load_this_plugin(self) -> ModuleTy:
        return self._load(str(self.originating_file), True)

    def load(self, path: str) -> ModuleTy:
        return self._load(path, False)

    def _load(self, path: str, is_root: bool) -> ModuleTy:
        resolved_path: pathlib.Path
        if is_root:
            resolved_path = pathlib.Path(path)
        else:
            resolved_path = paths.resolve_ruyi_load_path(
                path,
                self.root,
                False,
                self.originating_file,
                self.is_cmd,
            )
        resolved_path_str = str(resolved_path)
        if resolved_path_str in self.module_cache:
            return self.module_cache[resolved_path_str]

        plugin_id = resolved_path.relative_to(self.root).parts[0]
        plugin_dir = self.root / plugin_id

        host_bridge = api.make_ruyi_plugin_api_for_module(
            self._phctx,
            resolved_path,
            plugin_dir,
            self.is_cmd,
        )

        mod = self.do_load_module(
            resolved_path,
            resolved_path.read_text("utf-8"),
            host_bridge,
        )
        self.module_cache[resolved_path_str] = mod
        return mod

    @abc.abstractmethod
    def do_load_module(
        self,
        resolved_path: pathlib.Path,
        program: str,
        ruyi_host_bridge: Callable[[object], api.RuyiHostAPI],
    ) -> ModuleTy:
        raise NotImplementedError


# import the built-in supported PluginHostContext implementation(s)
# this must come after the baseclass declarations

# pylint: disable-next=wrong-import-position
from .unsandboxed import UnsandboxedPluginHostContext  # noqa: E402



import argparse
import pathlib
from typing import TYPE_CHECKING

from ...cli.cmd import RootCommand

if TYPE_CHECKING:
    from ...cli.completion import ArgumentParser
    from ...config import GlobalConfig


class VenvCommand(
    RootCommand,
    cmd="venv",
    help="Generate a virtual environment adapted to the chosen toolchain and profile",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument("profile", type=str, help="Profile to use for the environment")
        p.add_argument("dest", type=str, help="Path to the new virtual environment")
        p.add_argument(
            "--name",
            "-n",
            type=str,
            default=None,
            help="Override the venv's name",
        )
        p.add_argument(
            "--toolchain",
            "-t",
            type=str,
            action="append",
            help="Specifier(s) (atoms) of the toolchain package(s) to use",
        )
        p.add_argument(
            "--emulator",
            "-e",
            type=str,
            help="Specifier (atom) of the emulator package to use",
        )
        p.add_argument(
            "--with-sysroot",
            action="store_true",
            dest="with_sysroot",
            default=True,
            help="Provision a fresh sysroot inside the new virtual environment (default)",
        )
        p.add_argument(
            "--without-sysroot",
            action="store_false",
            dest="with_sysroot",
            help="Do not include a sysroot inside the new virtual environment",
        )
        p.add_argument(
            "--sysroot-from",
            type=str,
            help="Specifier (atom) of the sysroot package to use, in favor of the toolchain-included one if applicable",
        )
        p.add_argument(
            "--extra-commands-from",
            type=str,
            action="append",
            help="Specifier(s) (atoms) of extra package(s) to add commands to the new virtual environment",
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ...ruyipkg.host import get_native_host
        from .maker import do_make_venv

        profile_name: str = args.profile
        dest = pathlib.Path(args.dest)
        with_sysroot: bool = args.with_sysroot
        override_name: str | None = args.name
        tc_atoms_str: list[str] | None = args.toolchain
        emu_atom_str: str | None = args.emulator
        sysroot_atom_str: str | None = args.sysroot_from
        extra_cmd_atoms_str: list[str] | None = args.extra_commands_from
        host = str(get_native_host())

        return do_make_venv(
            cfg,
            host,
            profile_name,
            dest,
            with_sysroot,
            override_name,
            tc_atoms_str,
            emu_atom_str,
            sysroot_atom_str,
            extra_cmd_atoms_str,
        )



import argparse
from typing import TYPE_CHECKING

from .list_cli import ListCommand

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class ListProfilesCommand(
    ListCommand,
    cmd="profiles",
    help="List all available profiles",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        logger = cfg.logger
        mr = cfg.repo

        for arch in mr.get_supported_arches():
            for p in mr.iter_profiles_for_arch(arch):
                if not p.need_quirks:
                    logger.stdout(p.id)
                    continue

                logger.stdout(f"{p.id} (needs quirks: {p.need_quirks})")

        return 0



import argparse
import pathlib
from typing import TYPE_CHECKING, cast

from ..cli.cmd import AdminCommand

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class AdminChecksumCommand(
    AdminCommand,
    cmd="checksum",
    help="Generate a checksum section for a manifest file for the distfiles given",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--format",
            "-f",
            type=str,
            choices=["toml"],
            default="toml",
            help="Format of checksum section to generate in",
        )
        p.add_argument(
            "--restrict",
            type=str,
            default="",
            help="the 'restrict' field to use for all mentioned distfiles, separated with comma",
        )
        p.add_argument(
            "file",
            type=str,
            nargs="+",
            help="Path to the distfile(s) to checksum",
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .admin_checksum import do_admin_checksum

        logger = cfg.logger
        files = args.file
        format = args.format
        restrict_str = cast(str, args.restrict)
        restrict = restrict_str.split(",") if restrict_str else []

        return do_admin_checksum(logger, files, format, restrict)


class AdminFormatManifestCommand(
    AdminCommand,
    cmd="format-manifest",
    help="Format the given package manifests into canonical TOML representation",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "file",
            type=str,
            nargs="+",
            help="Path to the distfile(s) to generate manifest for",
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .canonical_dump import dumps_canonical_package_manifest_toml
        from .pkg_manifest import PackageManifest

        files = args.file

        for f in files:
            p = pathlib.Path(f)
            pm = PackageManifest.load_from_path(p)
            d = dumps_canonical_package_manifest_toml(pm)

            dest_path = p.with_suffix(".toml")
            with open(dest_path, "w", encoding="utf-8") as fp:
                fp.write(d)

        return 0



import calendar
import datetime
import functools
import json
import pathlib
import time
from typing import Callable, TYPE_CHECKING, cast
import uuid

from ..log import RuyiLogger
from ..utils.node_info import NodeInfo, gather_node_info
from .scope import TelemetryScope
from .store import TelemetryStore

if TYPE_CHECKING:
    # for avoiding circular import
    from ..config import GlobalConfig

FALLBACK_PM_TELEMETRY_ENDPOINT = "https://api.ruyisdk.cn/telemetry/pm/"

TELEMETRY_CONSENT_AND_UPLOAD_DESC = """
RuyiSDK collects anonymized usage data locally to help us improve the product.

[green]By default, nothing leaves your machine[/], and you can also turn off usage data
collection completely. Only with your explicit permission can [yellow]ruyi[/] upload
collected telemetry, periodically to RuyiSDK team-managed servers located in
the Chinese mainland. You can change this setting at any time by running
[yellow]ruyi telemetry consent[/], [yellow]ruyi telemetry local[/], or [yellow]ruyi telemetry optout[/].

If you enable uploads now, we'll also send a one-time report from this [yellow]ruyi[/]
installation so the RuyiSDK team can better understand adoption. Thank you for
helping us build a better experience!
"""
TELEMETRY_CONSENT_AND_UPLOAD_PROMPT = (
    "Enable telemetry uploads and send a one-time report now?"
)
TELEMETRY_OPTOUT_PROMPT = "\nDo you want to disable telemetry entirely?"


def next_utc_weekday(wday: int, now: float | None = None) -> int:
    t = time.gmtime(now)
    mday_delta = wday - t.tm_wday
    if mday_delta < 0:
        mday_delta += 7

    next_t = (
        t.tm_year,
        t.tm_mon,
        t.tm_mday + mday_delta,
        0,  # tm_hour
        0,  # tm_min
        0,  # tm_sec
        0,  # tm_wday
        0,  # tm_yday
        -1,  # tm_isdst
    )
    return calendar.timegm(next_t)


def set_telemetry_mode(
    gc: "GlobalConfig",
    mode: str,
    consent_time: datetime.datetime | None = None,
    show_cli_feedback: bool = True,
) -> None:
    """Set telemetry mode and consent time (if applicable) in the user preference."""

    from ..config.editor import ConfigEditor
    from ..config import schema

    logger = gc.logger

    if mode == "on":
        if consent_time is None:
            consent_time = datetime.datetime.now().astimezone()
    else:
        # clear any previously recorded consent time
        consent_time = None

    # First, persist the changes to user config
    with ConfigEditor.work_on_user_local_config(gc) as ed:
        ed.set_value((schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_MODE), mode)
        if consent_time is not None:
            ed.set_value(
                (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT),
                consent_time,
            )
        else:
            ed.unset_value(
                (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT)
            )

        ed.stage()

    # Then, apply the changes to the running instance's GlobalConfig
    # TelemetryProvider instance (if any) will pick them up automatically
    # because the properties are backed by GlobalConfig.
    gc.set_by_key(
        (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_MODE),
        mode,
    )
    gc.set_by_key(
        (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT),
        consent_time,
    )

    if not show_cli_feedback:
        return
    match mode:
        case "on":
            logger.I("telemetry data uploading is now enabled")
            logger.I(
                "you can opt out at any time by running [yellow]ruyi telemetry optout[/]"
            )
        case "local":
            logger.I("telemetry mode is now set to local collection only")
            logger.I(
                "you can re-enable telemetry data uploading at any time by running [yellow]ruyi telemetry consent[/]"
            )
            logger.I(
                "or opt out at any time by running [yellow]ruyi telemetry optout[/]"
            )
        case "off":
            logger.I("telemetry data collection is now disabled")
            logger.I(
                "you can re-enable telemetry data uploads at any time by running [yellow]ruyi telemetry consent[/]"
            )
        case _:
            raise ValueError(f"invalid telemetry mode: {mode}")


class TelemetryProvider:
    def __init__(self, gc: "GlobalConfig") -> None:
        self.state_root = pathlib.Path(gc.telemetry_root)

        self._discard_events = False
        self._gc = gc
        self._is_first_run = False
        self._stores: dict[TelemetryScope, TelemetryStore] = {}
        self._upload_on_exit = False

        # create the PM store
        self.init_store(TelemetryScope(None))
        # TODO: add real multi-repo support
        self.init_store(TelemetryScope("ruyisdk"))

    @property
    def logger(self) -> RuyiLogger:
        return self._gc.logger

    @property
    def local_mode(self) -> bool:
        return self._gc.telemetry_mode == "local"

    @property
    def upload_consent_time(self) -> datetime.datetime | None:
        return self._gc.telemetry_upload_consent_time

    def store(self, scope: TelemetryScope) -> TelemetryStore | None:
        return self._stores.get(scope)

    def init_store(self, scope: TelemetryScope) -> None:
        store_root = self.state_root
        api_url_fn: Callable[[], str | None] | None = None
        if repo_name := scope.repo_name:
            if repo_name != "ruyisdk":
                raise NotImplementedError("multi-repo support not implemented yet")
            store_root = store_root / "repos" / repo_name

            def _f() -> str | None:
                # access the repo attribute lazily to speed up CLI startup
                return self._gc.repo.get_telemetry_api_url("repo")

            api_url_fn = _f
        else:
            # configure the PM telemetry endpoint
            api_url_fn = functools.partial(self._detect_pm_api_url, self._gc)

        store = TelemetryStore(
            self.logger,
            scope,
            store_root,
            api_url_factory=api_url_fn,
        )
        self._stores[scope] = store

    def _detect_pm_api_url(self, gc: "GlobalConfig") -> str | None:
        url = FALLBACK_PM_TELEMETRY_ENDPOINT
        cfg_src = "fallback"
        if gc.override_pm_telemetry_url is not None:
            cfg_src = "local config"
            url = gc.override_pm_telemetry_url
        else:
            if repo_provided_url := gc.repo.get_telemetry_api_url("pm"):
                cfg_src = "repo"
                url = repo_provided_url
        self.logger.D(
            f"configured PM telemetry endpoint via {cfg_src}: {url or '(n/a)'}"
        )
        return url

    @property
    def installation_file(self) -> pathlib.Path:
        return self.state_root / "installation.json"

    def check_first_run_status(self) -> None:
        """Check if this is the first run of the application by checking if installation file exists.
        This must be done before init_installation() is potentially called.
        """
        self._is_first_run = not self.installation_file.exists()

    @property
    def is_first_run(self) -> bool:
        """Check if this is the first run of the application."""
        return self._is_first_run

    def init_installation(self, force_reinit: bool) -> NodeInfo | None:
        installation_file = self.installation_file
        if installation_file.exists() and not force_reinit:
            return self.read_installation_data()

        # either this is a fresh installation or we're forcing a refresh
        installation_id = uuid.uuid4()
        self.logger.D(
            f"initializing telemetry data store, installation_id={installation_id.hex}"
        )
        self.state_root.mkdir(parents=True, exist_ok=True)

        # (over)write installation data
        installation_data = gather_node_info(installation_id)
        with open(installation_file, "wb") as fp:
            fp.write(json.dumps(installation_data).encode("utf-8"))
        return installation_data

    def read_installation_data(self) -> NodeInfo | None:
        with open(self.installation_file, "rb") as fp:
            return cast(NodeInfo, json.load(fp))

    def upload_weekday(self) -> int | None:
        try:
            installation_data = self.read_installation_data()
        except FileNotFoundError:
            # init the node info if it's gone
            installation_data = self.init_installation(False)

        if installation_data is None:
            return None

        try:
            report_uuid_prefix = int(installation_data["report_uuid"][:8], 16)
        except ValueError:
            return None

        return report_uuid_prefix % 7  # 0 is Monday

    def has_upload_consent(self, time_now: float | None = None) -> bool:
        if self.upload_consent_time is None:
            return False
        if time_now is None:
            time_now = time.time()
        return self.upload_consent_time.timestamp() <= time_now

    def print_telemetry_notice(self, for_cli_verbose_output: bool = False) -> None:
        if self.local_mode:
            if for_cli_verbose_output:
                self.logger.I(
                    "telemetry mode is [green]local[/]: local data collection only, no uploads"
                )
            return

        now = time.time()
        if self.has_upload_consent(now) and not for_cli_verbose_output:
            self.logger.D("user has consented to telemetry upload")
            return

        upload_wday = self.upload_weekday()
        if upload_wday is None:
            return
        upload_wday_name = calendar.day_name[upload_wday]

        next_upload_day_ts = next_utc_weekday(upload_wday, now)
        next_upload_day = time.localtime(next_upload_day_ts)
        next_upload_day_end = time.localtime(next_upload_day_ts + 86400)
        next_upload_day_str = time.strftime("%Y-%m-%d %H:%M:%S %z", next_upload_day)
        next_upload_day_end_str = time.strftime(
            "%Y-%m-%d %H:%M:%S %z", next_upload_day_end
        )

        today_is_upload_day = self.is_upload_day(now)
        if for_cli_verbose_output:
            self.logger.I(
                "telemetry mode is [green]on[/]: data is collected and periodically uploaded"
            )
            self.logger.I(
                f"non-tracking usage information will be uploaded to RuyiSDK-managed servers [bold green]every {upload_wday_name}[/]"
            )
        else:
            self.logger.W(
                f"this [yellow]ruyi[/] installation has telemetry mode set to [yellow]on[/], and [bold]will upload non-tracking usage information to RuyiSDK-managed servers[/] [bold green]every {upload_wday_name}[/]"
            )

        if today_is_upload_day:
            for scope, store in self._stores.items():
                has_uploaded_today = self.has_uploaded_today(scope, now)
                if has_uploaded_today:
                    if last_upload_time := store.last_upload_timestamp:
                        last_upload_time_str = time.strftime(
                            "%Y-%m-%d %H:%M:%S %z", time.localtime(last_upload_time)
                        )
                        self.logger.I(
                            f"scope {scope}: usage information has already been uploaded today at {last_upload_time_str}"
                        )
                    else:
                        self.logger.I(
                            f"scope {scope}: usage information has already been uploaded sometime today"
                        )
                else:
                    self.logger.I(
                        f"scope {scope}: the next upload will happen [bold green]today[/] if not already"
                    )
        else:
            self.logger.I(
                f"the next upload will happen anytime [yellow]ruyi[/] is executed between [bold green]{next_upload_day_str}[/] and [bold green]{next_upload_day_end_str}[/]"
            )

        if not for_cli_verbose_output:
            self.logger.I("in order to hide this banner:")
            self.logger.I("- opt out with [yellow]ruyi telemetry optout[/]")
            self.logger.I("- or give consent with [yellow]ruyi telemetry consent[/]")

    def next_upload_day(self, time_now: float | None = None) -> int | None:
        upload_wday = self.upload_weekday()
        if upload_wday is None:
            return None
        return next_utc_weekday(upload_wday, time_now)

    def is_upload_day(self, time_now: float | None = None) -> bool:
        if time_now is None:
            time_now = time.time()
        if upload_day := self.next_upload_day(time_now):
            return upload_day <= time_now
        return False

    def has_uploaded_today(
        self,
        scope: TelemetryScope,
        time_now: float | None = None,
    ) -> bool:
        if time_now is None:
            time_now = time.time()
        if upload_day := self.next_upload_day(time_now):
            upload_day_end = upload_day + 86400
            store = self.store(scope)
            if store is None:
                return False
            if last_upload_time := store.last_upload_timestamp:
                return upload_day <= last_upload_time < upload_day_end
        return False

    def record(self, scope: TelemetryScope, kind: str, **params: object) -> None:
        if store := self.store(scope):
            return store.record(kind, **params)
        self.logger.D(f"no telemetry store for scope {scope}, discarding event")

    def discard_events(self, v: bool = True) -> None:
        self._discard_events = v

    def flush(self, *, upload_now: bool = False) -> None:
        now = time.time()

        # We may be self-uninstalling and purging all state data, and in this
        # case we don't want to record anything (thus re-creating directories).
        if self._discard_events:
            self.logger.D("discarding collected telemetry data")
            return

        self.logger.D("flushing telemetry to persistent store")

        for scope, store in self._stores.items():
            store.persist(now)

            # try to upload if forced (upload_now or _upload_on_exit), or:
            #
            # * we're not in local mode
            # * today is the day
            # * we haven't uploaded today
            if not (upload_now or self._upload_on_exit) and (
                self.local_mode
                or not self.is_upload_day(now)
                or self.has_uploaded_today(scope, now)
            ):
                continue

            self.prepare_data_for_upload(store)
            self.upload_staged_payloads(store)

    def prepare_data_for_upload(self, store: TelemetryStore) -> None:
        installation_data: NodeInfo | None = None
        if store.scope.is_pm:
            try:
                installation_data = self.read_installation_data()
            except FileNotFoundError:
                # should not happen due to is_upload_day() initializing it for us
                # beforehand, but proceed without node info nonetheless
                pass

        return store.prepare_data_for_upload(installation_data)

    def upload_staged_payloads(self, store: TelemetryStore) -> None:
        if self.local_mode:
            return

        return store.upload_staged_payloads()

    def oobe_prompt(self) -> None:
        """Ask whether the user consents to a first-run telemetry upload, and
        persist the user's exact telemetry choice."""

        from ..cli import user_input

        self.logger.stdout(TELEMETRY_CONSENT_AND_UPLOAD_DESC)
        if not user_input.ask_for_yesno_confirmation(
            self.logger,
            TELEMETRY_CONSENT_AND_UPLOAD_PROMPT,
            False,
        ):
            # ask if the user wants to opt out entirely
            if user_input.ask_for_yesno_confirmation(
                self.logger,
                TELEMETRY_OPTOUT_PROMPT,
                False,
            ):
                set_telemetry_mode(self._gc, "off")
                return

            # user wants to stay in local mode
            # explicitly record the preference, so we don't have to worry about
            # us potentially changing defaults yet another time
            set_telemetry_mode(self._gc, "local")
            return

        consent_time = datetime.datetime.now().astimezone()
        set_telemetry_mode(self._gc, "on", consent_time)
        self._upload_on_exit = True



import argparse
import datetime
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


# Telemetry preference commands
class TelemetryCommand(
    RootCommand,
    cmd="telemetry",
    has_subcommands=True,
    help="Manage your telemetry preferences",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass


class TelemetryConsentCommand(
    TelemetryCommand,
    cmd="consent",
    aliases=["on"],
    help="Give consent to telemetry data uploads",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provider import set_telemetry_mode

        now = datetime.datetime.now().astimezone()
        set_telemetry_mode(cfg, "on", now)
        return 0


class TelemetryLocalCommand(
    TelemetryCommand,
    cmd="local",
    help="Set telemetry mode to local collection only",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provider import set_telemetry_mode

        set_telemetry_mode(cfg, "local")
        return 0


class TelemetryOptoutCommand(
    TelemetryCommand,
    cmd="optout",
    aliases=["off"],
    help="Opt out of telemetry data collection",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provider import set_telemetry_mode

        set_telemetry_mode(cfg, "off")
        return 0


class TelemetryStatusCommand(
    TelemetryCommand,
    cmd="status",
    help="Print the current telemetry mode",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--verbose",
            "-v",
            action="store_true",
            help="Enable verbose output",
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        verbose: bool = args.verbose
        if not verbose:
            cfg.logger.stdout(cfg.telemetry_mode)
            return 0

        if cfg.telemetry is None:
            cfg.logger.I(
                "telemetry mode is [green]off[/]: no further data will be collected"
            )
            return 0

        cfg.telemetry.print_telemetry_notice(for_cli_verbose_output=True)
        return 0


class TelemetryUploadCommand(
    TelemetryCommand,
    cmd="upload",
    help="Upload collected telemetry data now",
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        if cfg.telemetry is None:
            cfg.logger.W("telemetry is disabled, nothing to upload")
            return 0

        cfg.telemetry.flush(upload_now=True)
        # disable the flush at program exit because we have just done that
        cfg.telemetry.discard_events()
        return 0



from functools import cached_property
import json
import os
import pathlib
import re
import time
from typing import Callable, Final, Iterable
import uuid

from ..log import RuyiLogger
from ..utils.node_info import NodeInfo
from ..utils.url import urljoin_for_sure
from .aggregate import UploadPayload, aggregate_events
from .event import TelemetryEvent, is_telemetry_event
from .scope import TelemetryScope

# e.g. "run.202410201845.d06ca5d668e64fec833ed3e6eb926a2c.ndjson"
RE_RAW_EVENT_FILENAME: Final = re.compile(
    r"^run\.(?P<time_bucket>\d{12})\.(?P<uuid>[0-9a-f]{32})\.ndjson$"
)


def get_time_bucket(timestamp: int | float | time.struct_time | None = None) -> str:
    if timestamp is None:
        return time.strftime("%Y%m%d%H%M")
    elif isinstance(timestamp, float) or isinstance(timestamp, int):
        timestamp = time.localtime(timestamp)
    return time.strftime("%Y%m%d%H%M", timestamp)


def time_bucket_from_filename(filename: str) -> str | None:
    if m := RE_RAW_EVENT_FILENAME.match(filename):
        return m.group("time_bucket")
    return None


class TelemetryStore:
    def __init__(
        self,
        logger: RuyiLogger,
        scope: TelemetryScope,
        store_root: pathlib.Path,
        api_url: str | None = None,
        api_url_factory: Callable[[], str | None] | None = None,
    ) -> None:
        self._logger = logger
        self.scope = scope
        self.store_root = store_root
        self._api_url = api_url
        self._api_url_factory = api_url_factory

        self._events: list[TelemetryEvent] = []

    @cached_property
    def api_url(self) -> str | None:
        if u := self._api_url:
            return u
        if f := self._api_url_factory:
            return f()
        return None

    @property
    def raw_events_dir(self) -> pathlib.Path:
        return self.store_root / "raw"

    @property
    def upload_stage_dir(self) -> pathlib.Path:
        return self.store_root / "staged"

    @property
    def uploaded_dir(self) -> pathlib.Path:
        return self.store_root / "uploaded"

    @property
    def last_upload_marker_file(self) -> pathlib.Path:
        return self.store_root / ".stamp-last-upload"

    @property
    def last_upload_timestamp(self) -> float | None:
        try:
            return self.last_upload_marker_file.stat().st_mtime
        except FileNotFoundError:
            return None

    def record_upload_timestamp(self, time_now: float | None = None) -> None:
        if time_now is None:
            time_now = time.time()
        f = self.last_upload_marker_file
        f.touch()
        os.utime(f, (time_now, time_now))

    def record(self, kind: str, **params: object) -> None:
        self._events.append({"fmt": 1, "kind": kind, "params": params})

    def discard_events(self, v: bool = True) -> None:
        self._discard_events = v

    def persist(self, now: float | None = None) -> None:
        if not self._events:
            self._logger.D(f"scope {self.scope}: no event to persist")
            return

        now = time.time() if now is None else now

        self._logger.D(f"scope {self.scope}: flushing telemetry to persistent store")

        raw_events_dir = self.raw_events_dir
        raw_events_dir.mkdir(parents=True, exist_ok=True)

        # TODO: for now it is safe to not lock, because flush() is only ever
        # called at program exit time
        rough_time = get_time_bucket(now)
        rand = uuid.uuid4().hex
        batch_events_file = raw_events_dir / f"run.{rough_time}.{rand}.ndjson"
        with open(batch_events_file, "wb") as fp:
            for e in self._events:
                payload = json.dumps(e)
                fp.write(payload.encode("utf-8"))
                fp.write(b"\n")

        self._logger.D(
            f"scope {self.scope}: persisted {len(self._events)} telemetry event(s)"
        )

    def upload(self, installation_data: NodeInfo | None = None) -> None:
        self.prepare_data_for_upload(installation_data)
        self.upload_staged_payloads()

    def read_back_raw_events(self) -> Iterable[TelemetryEvent]:
        try:
            for f in self.raw_events_dir.glob("run.*.ndjson"):
                time_bucket = time_bucket_from_filename(f.name)
                with open(f, "r", encoding="utf-8", newline=None) as fp:
                    for line in fp:
                        try:
                            obj = json.loads(line)
                        except json.JSONDecodeError:
                            # losing some malformed telemetry events is okay
                            continue
                        if not is_telemetry_event(obj):
                            # ditto
                            continue
                        if time_bucket is not None and "time_bucket" not in obj:
                            obj["time_bucket"] = time_bucket
                        yield obj
        except FileNotFoundError:
            pass

    def purge_raw_events(self) -> None:
        files = list(self.raw_events_dir.glob("run.*.ndjson"))
        for f in files:
            f.unlink(missing_ok=True)

    def gen_upload_staging_filename(self, nonce: str) -> pathlib.Path:
        return self.upload_stage_dir / f"staged.{nonce}.json"

    def prepare_data_for_upload(self, installation_data: NodeInfo | None) -> None:
        # import ruyi.version here because this package is on the CLI startup
        # critical path, and version probing is costly there
        from ..version import RUYI_SEMVER

        aggregate_data = list(aggregate_events(self.read_back_raw_events()))

        payload_nonce = uuid.uuid4().hex  # for server-side dedup purposes
        payload: UploadPayload = {
            "fmt": 1,
            "nonce": payload_nonce,
            "ruyi_version": RUYI_SEMVER,
            "events": aggregate_data,
        }
        if installation_data is not None:
            payload["installation"] = installation_data

        dest_path = self.gen_upload_staging_filename(payload_nonce)
        self.upload_stage_dir.mkdir(parents=True, exist_ok=True)
        dest_path.write_text(json.dumps(payload), encoding="utf-8")

        self.purge_raw_events()

    def upload_staged_payloads(self) -> None:
        if not self.api_url:
            return

        try:
            staged_payloads = list(self.upload_stage_dir.glob("staged.*.json"))
        except FileNotFoundError:
            return

        try:
            self.uploaded_dir.mkdir(parents=True, exist_ok=True)
        except OSError:
            return

        for f in staged_payloads:
            self.upload_one_staged_payload(f, self.api_url)

        self.record_upload_timestamp()

    def upload_one_staged_payload(
        self,
        f: pathlib.Path,
        endpoint: str,
    ) -> None:
        # import ruyi.version here because this package is on the CLI startup
        # critical path, and version probing is costly there
        from ..version import RUYI_USER_AGENT

        api_path = urljoin_for_sure(endpoint, "upload-v1")
        self._logger.D(f"scope {self.scope}: about to upload payload {f} to {api_path}")

        import requests

        resp = requests.post(
            api_path,
            data=f.read_bytes(),
            headers={"User-Agent": RUYI_USER_AGENT},
            allow_redirects=True,
            timeout=5,
        )

        if not (200 <= resp.status_code < 300):
            self._logger.D(
                f"scope {self.scope}: telemetry upload failed: status code {resp.status_code}, content {resp.content.decode('utf-8', 'replace')}"
            )
            return

        self._logger.D(
            f"scope {self.scope}: telemetry upload ok: status code {resp.status_code}"
        )

        # move to completed dir
        # TODO: rotation
        try:
            f.rename(self.uploaded_dir / f.name)
        except OSError as e:
            self._logger.D(
                f"scope {self.scope}: failed to move uploaded payload away: {e}"
            )
